@@ -21,7 +21,6 @@ import (
 	"golang.org/x/tools/gopls/internal/golang/completion/snippet"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
-	"golang.org/x/tools/internal/aliases"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/imports"
 	"golang.org/x/tools/internal/typesinternal"
@@ -70,7 +69,7 @@ type postfixTmplArgs struct {
 	// Type is the type of "foo.bar" in "foo.bar.print!".
 	Type types.Type
 
-	// FuncResult are results of the enclosed function
+	// FuncResults are results of the enclosed function
 	FuncResults []*types.Var
 
 	sel            *ast.SelectorExpr
@@ -78,7 +77,7 @@ type postfixTmplArgs struct {
 	snip           snippet.Builder
 	importIfNeeded func(pkgPath string, scope *types.Scope) (name string, edits []protocol.TextEdit, err error)
 	edits          []protocol.TextEdit
-	qf             types.Qualifier
+	qual           types.Qualifier
 	varNames       map[string]bool
 	placeholders   bool
 	currentTabStop int
@@ -102,11 +101,8 @@ var postfixTmpls = []postfixTmpl{{
 	label:   "reverse",
 	details: "reverse slice",
 	body: `{{if and (eq .Kind "slice") .StmtOK -}}
-{{$i := .VarName nil "i"}}{{$j := .VarName nil "j" -}}
-for {{$i}}, {{$j}} := 0, len({{.X}})-1; {{$i}} < {{$j}}; {{$i}}, {{$j}} = {{$i}}+1, {{$j}}-1 {
-	{{.X}}[{{$i}}], {{.X}}[{{$j}}] = {{.X}}[{{$j}}], {{.X}}[{{$i}}]
-}
-{{end}}`,
+{{.Import "slices"}}.Reverse({{.X}})
+{{- end}}`,
 }, {
 	label:   "range",
 	details: "range over slice",
@@ -390,7 +386,8 @@ func (a *postfixTmplArgs) EscapeQuotes(v string) string {
 
 // ElemType returns the Elem() type of xType, if applicable.
 func (a *postfixTmplArgs) ElemType() types.Type {
-	if e, _ := a.Type.(interface{ Elem() types.Type }); e != nil {
+	type hasElem interface{ Elem() types.Type } // Array, Chan, Map, Pointer, Slice
+	if e, ok := a.Type.Underlying().(hasElem); ok {
 		return e.Elem()
 	}
 	return nil
@@ -440,12 +437,13 @@ func (a *postfixTmplArgs) TypeName(t types.Type) (string, error) {
 	if t == nil || t == types.Typ[types.Invalid] {
 		return "", fmt.Errorf("invalid type: %v", t)
 	}
-	return types.TypeString(t, a.qf), nil
+	return types.TypeString(t, a.qual), nil
 }
 
 // Zero return the zero value representation of type t
 func (a *postfixTmplArgs) Zero(t types.Type) string {
-	return formatZeroValue(t, a.qf)
+	zero, _ := typesinternal.ZeroString(t, a.qual)
+	return zero
 }
 
 func (a *postfixTmplArgs) IsIdent() bool {
@@ -467,7 +465,7 @@ func (a *postfixTmplArgs) VarName(t types.Type, nonNamedDefault string) string {
 	// go/types predicates are undefined on types.Typ[types.Invalid].
 	if !types.Identical(t, types.Typ[types.Invalid]) && types.Implements(t, errorIntf) {
 		name = "err"
-	} else if !is[*types.Named](aliases.Unalias(typesinternal.Unpointer(t))) {
+	} else if !is[*types.Named](types.Unalias(typesinternal.Unpointer(t))) {
 		name = nonNamedDefault
 	}
 
@@ -506,13 +504,13 @@ func (c *completer) addPostfixSnippetCandidates(ctx context.Context, sel *ast.Se
 		return
 	}
 
-	selType := c.pkg.GetTypesInfo().TypeOf(sel.X)
+	selType := c.pkg.TypesInfo().TypeOf(sel.X)
 	if selType == nil {
 		return
 	}
 
 	// Skip empty tuples since there is no value to operate on.
-	if tuple, ok := selType.Underlying().(*types.Tuple); ok && tuple == nil {
+	if tuple, ok := selType.(*types.Tuple); ok && tuple == nil {
 		return
 	}
 
@@ -550,7 +548,7 @@ func (c *completer) addPostfixSnippetCandidates(ctx context.Context, sel *ast.Se
 		}
 	}
 
-	scope := c.pkg.GetTypes().Scope().Innermost(c.pos)
+	scope := c.pkg.Types().Scope().Innermost(c.pos)
 	if scope == nil {
 		return
 	}
@@ -586,11 +584,11 @@ func (c *completer) addPostfixSnippetCandidates(ctx context.Context, sel *ast.Se
 		tmplArgs := postfixTmplArgs{
 			X:              golang.FormatNode(c.pkg.FileSet(), sel.X),
 			StmtOK:         stmtOK,
-			Obj:            exprObj(c.pkg.GetTypesInfo(), sel.X),
+			Obj:            exprObj(c.pkg.TypesInfo(), sel.X),
 			Type:           selType,
 			FuncResults:    funcResults,
 			sel:            sel,
-			qf:             c.qf,
+			qual:           c.qual,
 			importIfNeeded: c.importIfNeeded,
 			scope:          scope,
 			varNames:       make(map[string]bool),
