@@ -7,14 +7,15 @@ package golang
 import (
 	"bytes"
 	"context"
-	"encoding/json"
 	"fmt"
 	"go/ast"
 	"go/constant"
 	"go/doc"
 	"go/format"
+	"go/printer"
 	"go/token"
 	"go/types"
+	"go/version"
 	"io/fs"
 	"path/filepath"
 	"sort"
@@ -36,7 +37,6 @@ import (
 	gastutil "golang.org/x/tools/gopls/internal/util/astutil"
 	"golang.org/x/tools/gopls/internal/util/bug"
 	"golang.org/x/tools/gopls/internal/util/safetoken"
-	"golang.org/x/tools/gopls/internal/util/typesutil"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/stdlib"
 	"golang.org/x/tools/internal/tokeninternal"
@@ -44,47 +44,41 @@ import (
 	"golang.org/x/tools/internal/typesinternal"
 )
 
-// hoverJSON contains the structured result of a hover query. It is
-// formatted in one of several formats as determined by the HoverKind
-// setting, one of which is JSON.
-//
-// We believe this is used only by govim.
-// TODO(adonovan): see if we can wean all clients of this interface.
-type hoverJSON struct {
-	// Synopsis is a single sentence synopsis of the symbol's documentation.
-	Synopsis string `json:"synopsis"`
+// hoverResult contains the (internal) result of a hover query.
+// It is formatted in one of several formats as determined by the
+// HoverKind setting.
+type hoverResult struct {
+	// synopsis is a single sentence synopsis of the symbol's documentation.
+	//
+	// TODO(adonovan): in what syntax? It (usually) comes from doc.synopsis,
+	// which produces "Text" form, but it may be fed to
+	// DocCommentToMarkdown, which expects doc comment syntax.
+	synopsis string
 
-	// FullDocumentation is the symbol's full documentation.
-	FullDocumentation string `json:"fullDocumentation"`
+	// fullDocumentation is the symbol's full documentation.
+	fullDocumentation string
 
-	// Signature is the symbol's signature.
-	Signature string `json:"signature"`
+	// signature is the symbol's signature.
+	signature string
 
-	// SingleLine is a single line describing the symbol.
+	// singleLine is a single line describing the symbol.
 	// This is recommended only for use in clients that show a single line for hover.
-	SingleLine string `json:"singleLine"`
+	singleLine string
 
-	// SymbolName is the human-readable name to use for the symbol in links.
-	SymbolName string `json:"symbolName"`
+	// symbolName is the human-readable name to use for the symbol in links.
+	symbolName string
 
-	// LinkPath is the pkg.go.dev link for the given symbol.
-	// For example, the "go/ast" part of "pkg.go.dev/go/ast#Node".
-	// It may have a module version suffix "@v1.2.3".
-	LinkPath string `json:"linkPath"`
+	// linkPath is the path of the package enclosing the given symbol,
+	// with the module portion (if any) replaced by "module@version".
+	//
+	// For example: "github.com/google/go-github/v48@v48.1.0/github".
+	//
+	// Use LinkTarget + "/" + linkPath + "#" + LinkAnchor to form a pkgsite URL.
+	linkPath string
 
-	// LinkAnchor is the pkg.go.dev link anchor for the given symbol.
+	// linkAnchor is the pkg.go.dev link anchor for the given symbol.
 	// For example, the "Node" part of "pkg.go.dev/go/ast#Node".
-	LinkAnchor string `json:"linkAnchor"`
-
-	// stdVersion is the Go release version at which this symbol became available.
-	// It is nil for non-std library.
-	stdVersion *stdlib.Version
-
-	// New fields go below, and are unexported. The existing
-	// exported fields are underspecified and have already
-	// constrained our movements too much. A detailed JSON
-	// interface might be nice, but it needs a design and a
-	// precise specification.
+	linkAnchor string
 
 	// typeDecl is the declaration syntax for a type,
 	// or "" for a non-type.
@@ -99,6 +93,10 @@ type hoverJSON struct {
 	// fields of a (struct) type that were promoted through an
 	// embedded field.
 	promotedFields string
+
+	// footer is additional content to insert at the bottom of the hover
+	// documentation, before the pkgdoc link.
+	footer string
 }
 
 // Hover implements the "textDocument/hover" RPC for Go files.
@@ -134,7 +132,7 @@ func Hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, positi
 // if the position is valid but we fail to compute hover information.
 //
 // TODO(adonovan): strength-reduce file.Handle to protocol.DocumentURI.
-func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp protocol.Position) (protocol.Range, *hoverJSON, error) {
+func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp protocol.Position) (protocol.Range, *hoverResult, error) {
 	// Check for hover inside the builtin file before attempting type checking
 	// below. NarrowestPackageForFile may or may not succeed, depending on
 	// whether this is a GOROOT view, but even if it does succeed the resulting
@@ -233,20 +231,25 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	// identifier.
 	for _, spec := range pgf.File.Imports {
 		if gastutil.NodeContains(spec, pos) {
-			rng, hoverJSON, err := hoverImport(ctx, snapshot, pkg, pgf, spec)
+			rng, hoverRes, err := hoverImport(ctx, snapshot, pkg, pgf, spec)
 			if err != nil {
 				return protocol.Range{}, nil, err
 			}
 			if hoverRange == nil {
 				hoverRange = &rng
 			}
-			return *hoverRange, hoverJSON, nil // (hoverJSON may be nil)
+			return *hoverRange, hoverRes, nil // (hoverRes may be nil)
 		}
 	}
-	// Handle hovering over (non-import-path) literals.
+
+	// Handle hovering over various special kinds of syntax node.
 	if path, _ := astutil.PathEnclosingInterval(pgf.File, pos, pos); len(path) > 0 {
-		if lit, _ := path[0].(*ast.BasicLit); lit != nil {
-			return hoverLit(pgf, lit, pos)
+		switch node := path[0].(type) {
+		// Handle hovering over (non-import-path) literals.
+		case *ast.BasicLit:
+			return hoverLit(pgf, node, pos)
+		case *ast.ReturnStmt:
+			return hoverReturnStatement(pgf, path, node)
 		}
 	}
 
@@ -270,7 +273,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 
 	// By convention, we qualify hover information relative to the package
 	// from which the request originated.
-	qf := typesutil.FileQualifier(pgf.File, pkg.Types(), pkg.TypesInfo())
+	qual := typesinternal.FileQualifier(pgf.File, pkg.Types())
 
 	// Handle type switch identifiers as a special case, since they don't have an
 	// object.
@@ -278,11 +281,11 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	// There's not much useful information to provide.
 	if selectedType != nil {
 		fakeObj := types.NewVar(obj.Pos(), obj.Pkg(), obj.Name(), selectedType)
-		signature := types.ObjectString(fakeObj, qf)
-		return *hoverRange, &hoverJSON{
-			Signature:  signature,
-			SingleLine: signature,
-			SymbolName: fakeObj.Name(),
+		signature := types.ObjectString(fakeObj, qual)
+		return *hoverRange, &hoverResult{
+			signature:  signature,
+			singleLine: signature,
+			symbolName: fakeObj.Name(),
 		}, nil
 	}
 
@@ -305,7 +308,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	docText := comment.Text()
 
 	// By default, types.ObjectString provides a reasonable signature.
-	signature := objectString(obj, qf, declPos, declPGF.Tok, spec)
+	signature := objectString(obj, qual, declPos, declPGF.Tok, spec)
 	singleLineSignature := signature
 
 	// Display struct tag for struct fields at the end of the signature.
@@ -316,7 +319,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	// TODO(rfindley): we could do much better for inferred signatures.
 	// TODO(adonovan): fuse the two calls below.
 	if inferred := inferredSignature(pkg.TypesInfo(), ident); inferred != nil {
-		if s := inferredSignatureString(obj, qf, inferred); s != "" {
+		if s := inferredSignatureString(obj, qual, inferred); s != "" {
 			signature = s
 		}
 	}
@@ -339,6 +342,13 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 	// if they embed platform-variant types.
 	//
 	var sizeOffset string // optional size/offset description
+	// debugging #69362: unexpected nil Defs[ident] value (?)
+	_ = ident.Pos()          // (can't be nil due to check after referencedObject)
+	_ = pkg.TypesInfo()      // (can't be nil due to check in call to inferredSignature)
+	_ = pkg.TypesInfo().Defs // (can't be nil due to nature of cache.Package)
+	if def, ok := pkg.TypesInfo().Defs[ident]; ok {
+		_ = def.Pos() // can't be nil due to reasoning in #69362.
+	}
 	if def, ok := pkg.TypesInfo().Defs[ident]; ok && ident.Pos() == def.Pos() {
 		// This is the declaring identifier.
 		// (We can't simply use ident.Pos() == obj.Pos() because
@@ -442,7 +452,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 			for _, f := range prom {
 				fmt.Fprintf(w, "%s\t%s\t// through %s\t\n",
 					f.field.Name(),
-					types.TypeString(f.field.Type(), qf),
+					types.TypeString(f.field.Type(), qual),
 					f.path)
 			}
 			w.Flush()
@@ -486,7 +496,7 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 			}
 
 			// Use objectString for its prettier rendering of method receivers.
-			b.WriteString(objectString(m.Obj(), qf, token.NoPos, nil, nil))
+			b.WriteString(objectString(m.Obj(), qual, token.NoPos, nil, nil))
 		}
 		methods = b.String()
 
@@ -598,37 +608,37 @@ func hover(ctx context.Context, snapshot *cache.Snapshot, fh file.Handle, pp pro
 		linkPath = strings.Replace(linkPath, mod.Path, mod.Path+"@"+mod.Version, 1)
 	}
 
-	var version *stdlib.Version
-	if symbol := StdSymbolOf(obj); symbol != nil {
-		version = &symbol.Version
+	var footer string
+	if sym := StdSymbolOf(obj); sym != nil && sym.Version > 0 {
+		footer = fmt.Sprintf("Added in %v", sym.Version)
 	}
 
-	return *hoverRange, &hoverJSON{
-		Synopsis:          doc.Synopsis(docText),
-		FullDocumentation: docText,
-		SingleLine:        singleLineSignature,
-		SymbolName:        linkName,
-		Signature:         signature,
-		LinkPath:          linkPath,
-		LinkAnchor:        anchor,
+	return *hoverRange, &hoverResult{
+		synopsis:          doc.Synopsis(docText),
+		fullDocumentation: docText,
+		singleLine:        singleLineSignature,
+		symbolName:        linkName,
+		signature:         signature,
+		linkPath:          linkPath,
+		linkAnchor:        anchor,
 		typeDecl:          typeDecl,
 		methods:           methods,
 		promotedFields:    fields,
-		stdVersion:        version,
+		footer:            footer,
 	}, nil
 }
 
 // hoverBuiltin computes hover information when hovering over a builtin
 // identifier.
-func hoverBuiltin(ctx context.Context, snapshot *cache.Snapshot, obj types.Object) (*hoverJSON, error) {
+func hoverBuiltin(ctx context.Context, snapshot *cache.Snapshot, obj types.Object) (*hoverResult, error) {
 	// Special handling for error.Error, which is the only builtin method.
 	//
 	// TODO(rfindley): can this be unified with the handling below?
 	if obj.Name() == "Error" {
 		signature := obj.String()
-		return &hoverJSON{
-			Signature:  signature,
-			SingleLine: signature,
+		return &hoverResult{
+			signature:  signature,
+			singleLine: signature,
 			// TODO(rfindley): these are better than the current behavior.
 			// SymbolName: "(error).Error",
 			// LinkPath:   "builtin",
@@ -670,14 +680,14 @@ func hoverBuiltin(ctx context.Context, snapshot *cache.Snapshot, obj types.Objec
 	signature = replacer.Replace(signature)
 
 	docText := comment.Text()
-	return &hoverJSON{
-		Synopsis:          doc.Synopsis(docText),
-		FullDocumentation: docText,
-		Signature:         signature,
-		SingleLine:        obj.String(),
-		SymbolName:        obj.Name(),
-		LinkPath:          "builtin",
-		LinkAnchor:        obj.Name(),
+	return &hoverResult{
+		synopsis:          doc.Synopsis(docText),
+		fullDocumentation: docText,
+		signature:         signature,
+		singleLine:        obj.String(),
+		symbolName:        obj.Name(),
+		linkPath:          "builtin",
+		linkAnchor:        obj.Name(),
 	}, nil
 }
 
@@ -685,7 +695,7 @@ func hoverBuiltin(ctx context.Context, snapshot *cache.Snapshot, obj types.Objec
 // imp in the file pgf of pkg.
 //
 // If we do not have metadata for the hovered import, it returns _
-func hoverImport(ctx context.Context, snapshot *cache.Snapshot, pkg *cache.Package, pgf *parsego.File, imp *ast.ImportSpec) (protocol.Range, *hoverJSON, error) {
+func hoverImport(ctx context.Context, snapshot *cache.Snapshot, pkg *cache.Package, pgf *parsego.File, imp *ast.ImportSpec) (protocol.Range, *hoverResult, error) {
 	rng, err := pgf.NodeRange(imp.Path)
 	if err != nil {
 		return protocol.Range{}, nil, err
@@ -728,15 +738,16 @@ func hoverImport(ctx context.Context, snapshot *cache.Snapshot, pkg *cache.Packa
 	}
 
 	docText := comment.Text()
-	return rng, &hoverJSON{
-		Synopsis:          doc.Synopsis(docText),
-		FullDocumentation: docText,
+	return rng, &hoverResult{
+		signature:         "package " + string(impMetadata.Name),
+		synopsis:          doc.Synopsis(docText),
+		fullDocumentation: docText,
 	}, nil
 }
 
 // hoverPackageName computes hover information for the package name of the file
 // pgf in pkg.
-func hoverPackageName(pkg *cache.Package, pgf *parsego.File) (protocol.Range, *hoverJSON, error) {
+func hoverPackageName(pkg *cache.Package, pgf *parsego.File) (protocol.Range, *hoverResult, error) {
 	var comment *ast.CommentGroup
 	for _, pgf := range pkg.CompiledGoFiles() {
 		if pgf.File.Doc != nil {
@@ -749,11 +760,47 @@ func hoverPackageName(pkg *cache.Package, pgf *parsego.File) (protocol.Range, *h
 		return protocol.Range{}, nil, err
 	}
 	docText := comment.Text()
-	return rng, &hoverJSON{
-		Synopsis:          doc.Synopsis(docText),
-		FullDocumentation: docText,
-		// Note: including a signature is redundant, since the cursor is already on the
-		// package name.
+
+	// List some package attributes at the bottom of the documentation, if
+	// applicable.
+	type attr struct{ title, value string }
+	var attrs []attr
+
+	if !metadata.IsCommandLineArguments(pkg.Metadata().ID) {
+		attrs = append(attrs, attr{"Package path", string(pkg.Metadata().PkgPath)})
+	}
+
+	if pkg.Metadata().Module != nil {
+		attrs = append(attrs, attr{"Module", pkg.Metadata().Module.Path})
+	}
+
+	// Show the effective language version for this package.
+	if v := pkg.TypesInfo().FileVersions[pgf.File]; v != "" {
+		attr := attr{value: version.Lang(v)}
+		if v == pkg.Types().GoVersion() {
+			attr.title = "Language version"
+		} else {
+			attr.title = "Language version (current file)"
+		}
+		attrs = append(attrs, attr)
+	}
+
+	// TODO(rfindley): consider exec'ing go here to compute DefaultGODEBUG, or
+	// propose adding GODEBUG info to go/packages.
+
+	var footer string
+	for i, attr := range attrs {
+		if i > 0 {
+			footer += "\n"
+		}
+		footer += fmt.Sprintf(" - %s: %s", attr.title, attr.value)
+	}
+
+	return rng, &hoverResult{
+		signature:         "package " + string(pkg.Metadata().Name),
+		synopsis:          doc.Synopsis(docText),
+		fullDocumentation: docText,
+		footer:            footer,
 	}, nil
 }
 
@@ -764,7 +811,7 @@ func hoverPackageName(pkg *cache.Package, pgf *parsego.File) (protocol.Range, *h
 // For example, hovering over "\u2211" in "foo \u2211 bar" yields:
 //
 //	'∑', U+2211, N-ARY SUMMATION
-func hoverLit(pgf *parsego.File, lit *ast.BasicLit, pos token.Pos) (protocol.Range, *hoverJSON, error) {
+func hoverLit(pgf *parsego.File, lit *ast.BasicLit, pos token.Pos) (protocol.Range, *hoverResult, error) {
 	var (
 		value      string    // if non-empty, a constant value to format in hover
 		r          rune      // if non-zero, format a description of this rune in hover
@@ -877,18 +924,57 @@ func hoverLit(pgf *parsego.File, lit *ast.BasicLit, pos token.Pos) (protocol.Ran
 		fmt.Fprintf(&b, "U+%04X, %s", r, runeName)
 	}
 	hover := b.String()
-	return rng, &hoverJSON{
-		Synopsis:          hover,
-		FullDocumentation: hover,
+	return rng, &hoverResult{
+		synopsis:          hover,
+		fullDocumentation: hover,
+	}, nil
+}
+
+func hoverReturnStatement(pgf *parsego.File, path []ast.Node, ret *ast.ReturnStmt) (protocol.Range, *hoverResult, error) {
+	var funcType *ast.FuncType
+	// Find innermost enclosing function.
+	for _, n := range path {
+		switch n := n.(type) {
+		case *ast.FuncLit:
+			funcType = n.Type
+		case *ast.FuncDecl:
+			funcType = n.Type
+		}
+		if funcType != nil {
+			break
+		}
+	}
+	// Inv: funcType != nil because a ReturnStmt is always enclosed by a function.
+	if funcType.Results == nil {
+		return protocol.Range{}, nil, nil // no result variables
+	}
+	rng, err := pgf.PosRange(ret.Pos(), ret.End())
+	if err != nil {
+		return protocol.Range{}, nil, err
+	}
+	// Format the function's result type.
+	var buf strings.Builder
+	var cfg printer.Config
+	fset := token.NewFileSet()
+	buf.WriteString("returns (")
+	for i, field := range funcType.Results.List {
+		if i > 0 {
+			buf.WriteString(", ")
+		}
+		cfg.Fprint(&buf, fset, field.Type)
+	}
+	buf.WriteByte(')')
+	return rng, &hoverResult{
+		signature: buf.String(),
 	}, nil
 }
 
 // hoverEmbed computes hover information for a filepath.Match pattern.
 // Assumes that the pattern is relative to the location of fh.
-func hoverEmbed(fh file.Handle, rng protocol.Range, pattern string) (protocol.Range, *hoverJSON, error) {
+func hoverEmbed(fh file.Handle, rng protocol.Range, pattern string) (protocol.Range, *hoverResult, error) {
 	s := &strings.Builder{}
 
-	dir := filepath.Dir(fh.URI().Path())
+	dir := fh.URI().DirPath()
 	var matches []string
 	err := filepath.WalkDir(dir, func(abs string, d fs.DirEntry, e error) error {
 		if e != nil {
@@ -917,30 +1003,30 @@ func hoverEmbed(fh file.Handle, rng protocol.Range, pattern string) (protocol.Ra
 		fmt.Fprintf(s, "%s\n\n", m)
 	}
 
-	json := &hoverJSON{
-		Signature:         fmt.Sprintf("Embedding %q", pattern),
-		Synopsis:          s.String(),
-		FullDocumentation: s.String(),
+	res := &hoverResult{
+		signature:         fmt.Sprintf("Embedding %q", pattern),
+		synopsis:          s.String(),
+		fullDocumentation: s.String(),
 	}
-	return rng, json, nil
+	return rng, res, nil
 }
 
 // inferredSignatureString is a wrapper around the types.ObjectString function
 // that adds more information to inferred signatures. It will return an empty string
 // if the passed types.Object is not a signature.
-func inferredSignatureString(obj types.Object, qf types.Qualifier, inferred *types.Signature) string {
+func inferredSignatureString(obj types.Object, qual types.Qualifier, inferred *types.Signature) string {
 	// If the signature type was inferred, prefer the inferred signature with a
 	// comment showing the generic signature.
 	if sig, _ := obj.Type().Underlying().(*types.Signature); sig != nil && sig.TypeParams().Len() > 0 && inferred != nil {
 		obj2 := types.NewFunc(obj.Pos(), obj.Pkg(), obj.Name(), inferred)
-		str := types.ObjectString(obj2, qf)
+		str := types.ObjectString(obj2, qual)
 		// Try to avoid overly long lines.
 		if len(str) > 60 {
 			str += "\n"
 		} else {
 			str += " "
 		}
-		str += "// " + types.TypeString(sig, qf)
+		str += "// " + types.TypeString(sig, qual)
 		return str
 	}
 	return ""
@@ -952,8 +1038,8 @@ func inferredSignatureString(obj types.Object, qf types.Qualifier, inferred *typ
 // syntax, and file must be the token.File describing its positions.
 //
 // Precondition: obj is not a built-in function or method.
-func objectString(obj types.Object, qf types.Qualifier, declPos token.Pos, file *token.File, spec ast.Spec) string {
-	str := types.ObjectString(obj, qf)
+func objectString(obj types.Object, qual types.Qualifier, declPos token.Pos, file *token.File, spec ast.Spec) string {
+	str := types.ObjectString(obj, qual)
 
 	switch obj := obj.(type) {
 	case *types.Func:
@@ -980,16 +1066,16 @@ func objectString(obj types.Object, qf types.Qualifier, declPos token.Pos, file
 					buf.WriteString(name)
 					buf.WriteString(" ")
 				}
-				types.WriteType(&buf, recv.Type(), qf)
+				types.WriteType(&buf, recv.Type(), qual)
 			}
 			buf.WriteByte(')')
 			buf.WriteByte(' ') // space (go/types uses a period)
-		} else if s := qf(obj.Pkg()); s != "" {
+		} else if s := qual(obj.Pkg()); s != "" {
 			buf.WriteString(s)
 			buf.WriteString(".")
 		}
 		buf.WriteString(obj.Name())
-		types.WriteSignature(&buf, sig, qf)
+		types.WriteSignature(&buf, sig, qual)
 		str = buf.String()
 
 	case *types.Const:
@@ -1144,9 +1230,10 @@ func parseFull(ctx context.Context, snapshot *cache.Snapshot, fset *token.FileSe
 }
 
 // If pkgURL is non-nil, it should be used to generate doc links.
-func formatHover(h *hoverJSON, options *settings.Options, pkgURL func(path PackagePath, fragment string) protocol.URI) (string, error) {
-	maybeMarkdown := func(s string) string {
-		if s != "" && options.PreferredContentFormat == protocol.Markdown {
+func formatHover(h *hoverResult, options *settings.Options, pkgURL func(path PackagePath, fragment string) protocol.URI) (string, error) {
+	markdown := options.PreferredContentFormat == protocol.Markdown
+	maybeFenced := func(s string) string {
+		if s != "" && markdown {
 			s = fmt.Sprintf("```go\n%s\n```", strings.Trim(s, "\n"))
 		}
 		return s
@@ -1154,54 +1241,75 @@ func formatHover(h *hoverJSON, options *settings.Options, pkgURL func(path Packa
 
 	switch options.HoverKind {
 	case settings.SingleLine:
-		return h.SingleLine, nil
+		return h.singleLine, nil
 
 	case settings.NoDocumentation:
-		return maybeMarkdown(h.Signature), nil
+		return maybeFenced(h.signature), nil
 
-	case settings.Structured:
-		b, err := json.Marshal(h)
-		if err != nil {
-			return "", err
-		}
-		return string(b), nil
+	case settings.SynopsisDocumentation, settings.FullDocumentation:
+		var sections [][]string // assembled below
 
-	case settings.SynopsisDocumentation,
-		settings.FullDocumentation:
+		// Signature section.
+		//
 		// For types, we display TypeDecl and Methods,
 		// but not Signature, which is redundant (= TypeDecl + "\n" + Methods).
 		// For all other symbols, we display Signature;
 		// TypeDecl and Methods are empty.
-		// (This awkwardness is to preserve JSON compatibility.)
-		parts := []string{
-			maybeMarkdown(h.Signature),
-			maybeMarkdown(h.typeDecl),
-			formatDoc(h, options),
-			maybeMarkdown(h.promotedFields),
-			maybeMarkdown(h.methods),
-			fmt.Sprintf("Added in %v", h.stdVersion),
-			formatLink(h, options, pkgURL),
-		}
+		// (Now that JSON is no more, we could rationalize this.)
 		if h.typeDecl != "" {
-			parts[0] = "" // type: suppress redundant Signature
-		}
-		if h.stdVersion == nil || *h.stdVersion == stdlib.Version(0) {
-			parts[5] = "" // suppress stdlib version if not applicable or initial version 1.0
+			sections = append(sections, []string{maybeFenced(h.typeDecl)})
+		} else {
+			sections = append(sections, []string{maybeFenced(h.signature)})
 		}
 
+		// Doc section.
+		var doc string
+		switch options.HoverKind {
+		case settings.SynopsisDocumentation:
+			doc = h.synopsis
+		case settings.FullDocumentation:
+			doc = h.fullDocumentation
+		}
+		if options.PreferredContentFormat == protocol.Markdown {
+			doc = DocCommentToMarkdown(doc, options)
+		}
+		sections = append(sections, []string{
+			doc,
+			maybeFenced(h.promotedFields),
+			maybeFenced(h.methods),
+		})
+
+		// Footer section.
+		sections = append(sections, []string{
+			h.footer,
+			formatLink(h, options, pkgURL),
+		})
+
 		var b strings.Builder
-		for _, part := range parts {
-			if part == "" {
-				continue
+		newline := func() {
+			if options.PreferredContentFormat == protocol.Markdown {
+				b.WriteString("\n\n")
+			} else {
+				b.WriteByte('\n')
 			}
-			if b.Len() > 0 {
-				if options.PreferredContentFormat == protocol.Markdown {
-					b.WriteString("\n\n")
-				} else {
-					b.WriteByte('\n')
+		}
+		for _, section := range sections {
+			start := b.Len()
+			for _, part := range section {
+				if part == "" {
+					continue
 				}
+				// When markdown is a available, insert an hline before the start of
+				// the section, if there is content above.
+				if markdown && b.Len() == start && start > 0 {
+					newline()
+					b.WriteString("---")
+				}
+				if b.Len() > 0 {
+					newline()
+				}
+				b.WriteString(part)
 			}
-			b.WriteString(part)
 		}
 		return b.String(), nil
 
@@ -1240,7 +1348,7 @@ func StdSymbolOf(obj types.Object) *stdlib.Symbol {
 	// Handle Method.
 	if fn, _ := obj.(*types.Func); fn != nil {
 		isPtr, named := typesinternal.ReceiverNamed(fn.Signature().Recv())
-		if isPackageLevel(named.Obj()) {
+		if named != nil && isPackageLevel(named.Obj()) {
 			for _, s := range symbols {
 				if s.Kind != stdlib.Method {
 					continue
@@ -1282,26 +1390,35 @@ func StdSymbolOf(obj types.Object) *stdlib.Symbol {
 }
 
 // If pkgURL is non-nil, it should be used to generate doc links.
-func formatLink(h *hoverJSON, options *settings.Options, pkgURL func(path PackagePath, fragment string) protocol.URI) string {
-	if options.LinksInHover == settings.LinksInHover_None || h.LinkPath == "" {
+func formatLink(h *hoverResult, options *settings.Options, pkgURL func(path PackagePath, fragment string) protocol.URI) string {
+	if options.LinksInHover == settings.LinksInHover_None || h.linkPath == "" {
 		return ""
 	}
 	var url protocol.URI
 	var caption string
 	if pkgURL != nil { // LinksInHover == "gopls"
-		path, _, _ := strings.Cut(h.LinkPath, "@") // remove optional module version suffix
-		url = pkgURL(PackagePath(path), h.LinkAnchor)
+		// Discard optional module version portion.
+		// (Ideally the hoverResult would retain the structure...)
+		path := h.linkPath
+		if module, versionDir, ok := strings.Cut(h.linkPath, "@"); ok {
+			// "module@version/dir"
+			path = module
+			if _, dir, ok := strings.Cut(versionDir, "/"); ok {
+				path += "/" + dir
+			}
+		}
+		url = pkgURL(PackagePath(path), h.linkAnchor)
 		caption = "in gopls doc viewer"
 	} else {
 		if options.LinkTarget == "" {
 			return ""
 		}
-		url = cache.BuildLink(options.LinkTarget, h.LinkPath, h.LinkAnchor)
+		url = cache.BuildLink(options.LinkTarget, h.linkPath, h.linkAnchor)
 		caption = "on " + options.LinkTarget
 	}
 	switch options.PreferredContentFormat {
 	case protocol.Markdown:
-		return fmt.Sprintf("[`%s` %s](%s)", h.SymbolName, caption, url)
+		return fmt.Sprintf("[`%s` %s](%s)", h.symbolName, caption, url)
 	case protocol.PlainText:
 		return ""
 	default:
@@ -1309,20 +1426,6 @@ func formatLink(h *hoverJSON, options *settings.Options, pkgURL func(path Packag
 	}
 }
 
-func formatDoc(h *hoverJSON, options *settings.Options) string {
-	var doc string
-	switch options.HoverKind {
-	case settings.SynopsisDocumentation:
-		doc = h.Synopsis
-	case settings.FullDocumentation:
-		doc = h.FullDocumentation
-	}
-	if options.PreferredContentFormat == protocol.Markdown {
-		return CommentToMarkdown(doc, options)
-	}
-	return doc
-}
-
 // findDeclInfo returns the syntax nodes involved in the declaration of the
 // types.Object with position pos, searching the given list of file syntax
 // trees.
@@ -1563,7 +1666,12 @@ func computeSizeOffsetInfo(pkg *cache.Package, path []ast.Node, obj types.Object
 		var tStruct *types.Struct
 		for _, n := range path {
 			if n, ok := n.(*ast.StructType); ok {
-				tStruct = pkg.TypesInfo().TypeOf(n).(*types.Struct)
+				t, ok := pkg.TypesInfo().TypeOf(n).(*types.Struct)
+				if ok {
+					// golang/go#69150: TypeOf(n) was observed not to be a Struct (likely
+					// nil) in some cases.
+					tStruct = t
+				}
 				break
 			}
 		}
