@@ -132,18 +132,16 @@ type Interface interface {
 	// Runs `go get` to fetch a package.
 	GoGetPackage(context.Context, GoGetPackageArgs) error
 
-	// GCDetails: Toggle gc_details
+	// GCDetails: Toggle display of compiler optimization details
 	//
-	// Toggle the calculation of gc annotations.
+	// Toggle the per-package flag that causes Go compiler
+	// optimization decisions to be reported as diagnostics.
+	//
+	// (The name is a legacy of a time when the Go compiler was
+	// known as "gc". Renaming the command would break custom
+	// client-side logic in VS Code.)
 	GCDetails(context.Context, protocol.DocumentURI) error
 
-	// TODO: deprecate GCDetails in favor of ToggleGCDetails below.
-
-	// ToggleGCDetails: Toggle gc_details
-	//
-	// Toggle the calculation of gc annotations.
-	ToggleGCDetails(context.Context, URIArg) error
-
 	// ListKnownPackages: List known packages
 	//
 	// Retrieve a list of packages that are importable from the given URI.
@@ -525,7 +523,7 @@ type RunVulncheckResult struct {
 	Token protocol.ProgressToken
 }
 
-// GovulncheckResult holds the result of synchronously running the vulncheck
+// VulncheckResult holds the result of synchronously running the vulncheck
 // command.
 type VulncheckResult struct {
 	// Result holds the result of running vulncheck.
@@ -695,7 +693,7 @@ type PackagesResult struct {
 	// Packages is an unordered list of package metadata.
 	Packages []Package
 
-	// Modules maps module path to module metadata for
+	// Module maps module path to module metadata for
 	// all the modules of the returned Packages.
 	Module map[string]Module
 }
