@@ -61,9 +61,12 @@ func DefaultOptions(overrides ...func(*Options)) *Options {
 						RefactorRewriteRemoveUnusedParam: true,
 						RefactorRewriteSplitLines:        true,
 						RefactorInlineCall:               true,
+						RefactorExtractConstant:          true,
+						RefactorExtractConstantAll:       true,
 						RefactorExtractFunction:          true,
 						RefactorExtractMethod:            true,
 						RefactorExtractVariable:          true,
+						RefactorExtractVariableAll:       true,
 						RefactorExtractToNewFile:         true,
 						// Not GoTest: it must be explicit in CodeActionParams.Context.Only
 					},
@@ -86,12 +89,6 @@ func DefaultOptions(overrides ...func(*Options)) *Options {
 				},
 				UIOptions: UIOptions{
 					DiagnosticOptions: DiagnosticOptions{
-						Annotations: map[Annotation]bool{
-							Bounds: true,
-							Escape: true,
-							Inline: true,
-							Nil:    true,
-						},
 						Vulncheck:                 ModeVulncheckOff,
 						DiagnosticsDelay:          1 * time.Second,
 						DiagnosticsTrigger:        DiagnosticsOnEdit,
@@ -119,7 +116,6 @@ func DefaultOptions(overrides ...func(*Options)) *Options {
 						CodeLensGenerate:          true,
 						CodeLensRegenerateCgo:     true,
 						CodeLensTidy:              true,
-						CodeLensGCDetails:         false,
 						CodeLensUpgradeDependency: true,
 						CodeLensVendor:            true,
 						CodeLensRunGovulncheck:    false, // TODO(hyangah): enable
