@@ -43,7 +43,13 @@ func TestDiagnosticErrorInEditedFile(t *testing.T) {
 	// This test is very basic: start with a clean Go program, make an error, and
 	// get a diagnostic for that error. However, it also demonstrates how to
 	// combine Expectations to await more complex state in the editor.
-	Run(t, exampleProgram, func(t *testing.T, env *Env) {
+	RunMultiple{
+		{"golist", WithOptions(Modes(Default))},
+		{"gopackages", WithOptions(
+			Modes(Default),
+			FakeGoPackagesDriver(t),
+		)},
+	}.Run(t, exampleProgram, func(t *testing.T, env *Env) {
 		// Deleting the 'n' at the end of Println should generate a single error
 		// diagnostic.
 		env.OpenFile("main.go")
@@ -66,15 +72,22 @@ module mod.com
 
 go 1.12
 `
-	Run(t, onlyMod, func(t *testing.T, env *Env) {
+	WithOptions(
+		Settings{
+			"pullDiagnostics": true,
+		},
+	).Run(t, onlyMod, func(t *testing.T, env *Env) {
 		env.CreateBuffer("main.go", `package main
 
-func m() {
+func _() {
 	log.Println()
 }
 `)
 		env.AfterChange(Diagnostics(env.AtRegexp("main.go", "log")))
 		env.SaveBuffer("main.go")
+		if got := env.Diagnostics("main.go"); len(got) != 0 {
+			t.Errorf("got %d diagnostics, want 0", len(got))
+		}
 		env.AfterChange(NoDiagnostics(ForFile("main.go")))
 	})
 }
@@ -84,7 +97,15 @@ func TestDiagnosticErrorInNewFile(t *testing.T) {
 
 const Foo = "abc
 `
-	Run(t, brokenFile, func(t *testing.T, env *Env) {
+	RunMultiple{
+		{"golist", WithOptions(Modes(Default))},
+		// Since this test requires loading an overlay,
+		// it verifies that the fake go/packages driver honors overlays.
+		{"gopackages", WithOptions(
+			Modes(Default),
+			FakeGoPackagesDriver(t),
+		)},
+	}.Run(t, brokenFile, func(t *testing.T, env *Env) {
 		env.CreateBuffer("broken.go", brokenFile)
 		env.AfterChange(Diagnostics(env.AtRegexp("broken.go", "\"abc")))
 	})
@@ -107,8 +128,18 @@ const a = 2
 `
 
 func TestDiagnosticClearingOnEdit(t *testing.T) {
-	Run(t, badPackage, func(t *testing.T, env *Env) {
+	WithOptions(
+		Settings{
+			"pullDiagnostics": true,
+		},
+	).Run(t, badPackage, func(t *testing.T, env *Env) {
 		env.OpenFile("b.go")
+
+		for _, f := range []string{"a.go", "b.go"} {
+			if got := env.Diagnostics(f); len(got) != 1 {
+				t.Errorf("textDocument/diagnostic(%s) returned %d diagnostics, want 1. Got %v", f, len(got), got)
+			}
+		}
 		env.AfterChange(
 			Diagnostics(env.AtRegexp("a.go", "a = 1")),
 			Diagnostics(env.AtRegexp("b.go", "a = 2")),
@@ -116,6 +147,11 @@ func TestDiagnosticClearingOnEdit(t *testing.T) {
 
 		// Fix the error by editing the const name in b.go to `b`.
 		env.RegexpReplace("b.go", "(a) = 2", "b")
+		for _, f := range []string{"a.go", "b.go"} {
+			if got := env.Diagnostics(f); len(got) != 0 {
+				t.Errorf("textDocument/diagnostic(%s) returned %d diagnostics, want 0. Got %v", f, len(got), got)
+			}
+		}
 		env.AfterChange(
 			NoDiagnostics(ForFile("a.go")),
 			NoDiagnostics(ForFile("b.go")),
@@ -299,7 +335,7 @@ func Hello() {
 				InitialWorkspaceLoad,
 				Diagnostics(env.AtRegexp("main.go", `"mod.com/bob"`)),
 			)
-			if err := env.Sandbox.RunGoCommand(env.Ctx, "", "mod", []string{"init", "mod.com"}, nil, true); err != nil {
+			if _, err := env.Sandbox.RunGoCommand(env.Ctx, "", "mod", []string{"init", "mod.com"}, nil, true); err != nil {
 				t.Fatal(err)
 			}
 			env.AfterChange(
@@ -559,7 +595,7 @@ func f() {
 			NoOutstandingWork(IgnoreTelemetryPromptWork),
 			Diagnostics(
 				env.AtRegexp("a.go", `"mod.com`),
-				WithMessage("GOROOT or GOPATH"),
+				WithMessage("in GOROOT"),
 			),
 		)
 		// Deleting the import dismisses the warning.
